/*
 *
  ***** BEGIN LICENSE BLOCK *****
 
  Copyright (C) 2009-2018 Olof Hagsand and Benny Holmgren

  This file is part of CLIXON.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  Alternatively, the contents of this file may be used under the terms of
  the GNU General Public License Version 3 or later (the "GPL"),
  in which case the provisions of the GPL are applicable instead
  of those above. If you wish to allow use of your version of this file only
  under the terms of the GPL, and not to allow others to
  use your version of this file under the terms of Apache License version 2, 
  indicate your decision by deleting the provisions above and replace them with
  the  notice and other provisions required by the GPL. If you do not delete
  the provisions above, a recipient may use your version of this file under
  the terms of any one of the Apache License version 2 or the GPL.

  ***** END LICENSE BLOCK *****

See https://www.w3.org/TR/xpath/

 * Turn this on to get an xpath test program 
 * Usage: xpath [<xpath>] 
 * read xpath on first line and xml on rest of lines from input
 * Example compile:
 gcc -g -o xpath -I. -I../clixon ./clixon_xsl.c -lclixon -lcligen
 * Example run:
echo "a\n<a><b/></a>" | xpath
*/

#ifdef HAVE_CONFIG_H
#include "clixon_config.h" /* generated by config & autoconf */
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <limits.h>
#include <fnmatch.h>
#include <stdint.h>
#include <assert.h>
#include <syslog.h>
#include <fcntl.h>

/* cligen */
#include <cligen/cligen.h>

/* clixon */
#include "clixon/clixon.h"

static int
usage(char *argv0)
{
    fprintf(stderr, "usage:%s [options]\n"
	    "where options are\n"
            "\t-h \t\tHelp\n"
    	    "\t-D <level> \tDebug\n"
	    "\t-f <file>  \tXML file\n"
	    "\t-p <xpath> \tPrimary XPATH string\n"
	    "\t-i <xpath0>\t(optional) Initial XPATH string\n"
	    "and the following extra rules:\n"
	    "\tif -f is not given, XML input is expected on stdin\n"
	    "\tif -p is not given, <xpath> is expected as the first line on stdin\n"
	    "This means that with no arguments, <xpath> and XML is expected on stadin.\n",
	    argv0
	    );
    exit(0);
}

static int
ctx_print2(cbuf   *cb,
	   xp_ctx *xc)
{
    int        i;

    cprintf(cb, "%s:", (char*)clicon_int2str(ctxmap, xc->xc_type));
    switch (xc->xc_type){
    case XT_NODESET:
	for (i=0; i<xc->xc_size; i++){
	    cprintf(cb, "%d:", i);
	    clicon_xml2cbuf(cb, xc->xc_nodeset[i], 0, 0);
	}
	break;
    case XT_BOOL:
	cprintf(cb, "%s", xc->xc_bool?"true":"false");
	break;
    case XT_NUMBER:
	cprintf(cb, "%lf", xc->xc_number);
	break;
    case XT_STRING:
	cprintf(cb, "%s", xc->xc_string);
	break;
    }
    return 0;
}

int
main(int argc, char **argv)
{
    int         retval = -1;
    char       *argv0 = argv[0];
    int         i;
    cxobj     **xv = NULL;
    cxobj      *x0 = NULL;
    cxobj      *x;
    char        c;
    int         len;
    char       *buf = NULL;
    int         ret;
    int         fd = 0; /* unless overriden by argv[1] */
    char       *xpath = NULL;
    char       *xpath0 = NULL;
    char       *filename;
    xp_ctx     *xc = NULL;
    cbuf       *cb = NULL;
    
    clicon_log_init("xpath", LOG_DEBUG, CLICON_LOG_STDERR); 
    optind = 1;
    opterr = 0;
    while ((c = getopt(argc, argv, "hDf:p:i:")) != -1)
	switch (c) {
	case 'h':
	    usage(argv0);
	    break;
    	case 'D':
	    debug++;
	    break;
	case 'f': /* XML file */
	    filename = optarg;
	    if ((fd = open(filename, O_RDONLY)) < 0){
		clicon_err(OE_UNIX, errno, "open(%s)", argv[1]);
		goto done;
	    }
	    break;
	case 'p': /* Primary XPATH string */
	    xpath = optarg;
	    break;
	case 'i': /* Optional initial XPATH string */
	    xpath0 = optarg;
	    break;
	default:
	    usage(argv[0]);
	    break;
	}
    if (xpath==NULL){
	/* First read xpath */
	len = 1024; /* any number is fine */
	if ((buf = malloc(len)) == NULL){
	    perror("pt_file malloc");
	    return -1;
	}
	memset(buf, 0, len);
	i = 0;
	while (1){ 
	    if ((ret = read(0, &c, 1)) < 0){
		perror("read");
		goto done;
	    }
	    if (ret == 0)
		break;
	    if (c == '\n')
		break;
	    if (len==i){
		if ((buf = realloc(buf, 2*len)) == NULL){
		    fprintf(stderr, "%s: realloc: %s\n", __FUNCTION__, strerror(errno));
		    return -1;
		}	    
		memset(buf+len, 0, len);
		len *= 2;
	    }
	    buf[i++] = (char)(c&0xff);
	}
	xpath = buf;
    }
    /* 
     * If fd=0, then continue reading from stdin (after CR)
     * If fd>0, reading from file opened as argv[1]
     */
    if (xml_parse_file(fd, "</clicon>", NULL, &x0) < 0){
	fprintf(stderr, "Error: parsing: %s\n", clicon_err_reason);
	return -1;
    }
    /* If xpath0 given, position current x */
    if (xpath0){
	if ((x = xpath_first(x0, "%s", xpath0)) == NULL){
	    fprintf(stderr, "Error: xpath0 returned NULL\n");
	    return -1;
	}
    }
    else
	x = x0;

    /* Parse XML */
    if (xpath_vec_ctx(x, xpath, &xc) < 0)
	return -1;
    /* Print results */
    cb = cbuf_new();
    ctx_print2(cb, xc);
    fprintf(stdout, "%s\n", cbuf_get(cb));
    retval = 0;
 done:
    if (cb)
	cbuf_free(cb);
    if (xc)
	ctx_free(xc);
    if (xv)
	free(xv);
    if (buf)
	free(buf);
    if (x0)
	xml_free(x0);
    if (fd > 0)
	close(fd);
    return retval;
}
